var gColorPalette = [];
var gPref = Components.classes["@mozilla.org/preferences-service;1"]
                      .getService(Components.interfaces.nsIPrefService)
                      .getBranch(null);
var gStartTime = 0;

function Startup() {
  updateIterMenu();
  updatePaletteMenu();
  document.getElementById("statusLabel").value =
      document.getElementById("mbrotBundle").getString("statusEmpty");
}

function drawImage() {
  let canvas = document.getElementById("mbrotImage");
  if (canvas.getContext) {
    let context = canvas.getContext("2d");

    document.getElementById("statusLabel").value =
        document.getElementById("mbrotBundle").getString("statusDrawing");

    // example:
    // context.fillStyle = "rgb(200,0,0)";
    // context.fillRect (10, 10, 55, 50); // x, y, width, height
    //
    // context.fillStyle = "rgba(0, 0, 200, 0.5)";
    // context.fillRect (30, 30, 55, 50);

    let Cr_min = -2.0;
    let Cr_max = 1.0;
    let Cr_scale = Cr_max - Cr_min;

    let Ci_min = -1.5;
    let Ci_max = 1.5;
    let Ci_scale = Ci_max - Ci_min;

    let iterMax = gPref.getIntPref("mandelbrot.iteration_max");
    let algorithm = gPref.getCharPref("mandelbrot.use_algorithm");

    gStartTime = new Date();

    for (let img_x = 0; img_x < canvas.width; img_x++) {
      for (let img_y = 0; img_y < canvas.height; img_y++) {
        let C = new complex(Cr_min + (img_x / canvas.width) * Cr_scale,
                            Ci_min + (img_y / canvas.height) * Ci_scale);
        window.setTimeout(drawPoint, 0, context, img_x, img_y, C, iterMax, algorithm);
      }
    }
    window.setTimeout(EndCalc, 0);
  }
}

function EndCalc() {
  let endTime = new Date();
  let timeUsed = (endTime.getTime() - gStartTime.getTime()) / 1000;
  document.getElementById("statusLabel").value =
      document.getElementById("mbrotBundle").getFormattedString("statusTime", [timeUsed.toFixed(3)]);
}

function complex(aReal, aImag) {
  this.r = aReal;
  this.i = aImag;
  this.square = function() {
    return new complex(this.r * this.r - this.i * this.i,
                       2 * this.r * this.i);
  }
  this.dist = function() {
    return Math.sqrt(this.r * this.r + this.i * this.i);
  }
  this.add = function(aComplex) {
    return new complex(this.r + aComplex.r, this.i + aComplex.i);
  }
}

function mandelbrotValueOO (aC, aIterMax) {
  // this would be nice code in general but it looks like JS objects are too heavy for normal use.
  let Z = new complex(0.0, 0.0);
  for (var iter = 0; iter < aIterMax; iter++) {
    Z = Z.square().add(aC);
    if (Z.r * Z.r + Z.i * Z.i > 256) { break; }
  }
  return iter;
}

function mandelbrotValueNumeric (aC, aIterMax) {
  // optimized numeric code for fast calculation
  let Cr = aC.r, Ci = aC.i;
  let Zr = 0.0, Zi = 0.0;
  let Zr2 = Zr * Zr, Zi2 = Zi * Zi;
  for (var iter = 0; iter < aIterMax; iter++) {
    Zi = 2 * Zr * Zi + Ci;
    Zr = Zr2 - Zi2 + Cr;

    Zr2 = Zr * Zr; Zi2 = Zi * Zi;
    if (Zr2 + Zi2 > 256) { break; }
  }
  return iter;
}

function getColor(aIterValue, aIterMax) {
  let standardizedValue = Math.round(aIterValue * 1024 / aIterMax);
  if (gColorPalette && gColorPalette.length)
    return gColorPalette[standardizedValue];

  // fallback to simple b/w if for some reason we don't have a palette
  if (aIterValue == aIterMax)
    return "rgb(0,0,0)";
  else
    return "rgb(255,255,255)";
}

function getColorPalette(palName) {
  var palette = [];
  switch (palName) {
    case 'bw':
      for (let i = 0; i < 1024; i++) {
        palette[i] = 'rgb(255,255,255)';
      }
      palette[1024] = 'rgb(0,0,0)';
      break;
    case 'kairo':
      // outer areas
      for (let i = 0; i < 32; i++) {
        let cc1 = Math.floor(i * 127 / 31);
        let cc2 = 170 - Math.floor(i * 43 / 31);
        palette[i] = 'rgb(' + cc1 + ',' + cc2 + ',' + cc1 + ')';
      }
      // inner areas
      for (let i = 0; i < 51; i++) {
        let cc = Math.floor(i * 170 / 50);
        palette[32 + i] = 'rgb(' + cc + ',0,' + (170 + cc) + ')';
      }
      // corona
      for (let i = 0; i < 101; i++) {
        let cc = Math.floor(i * 200 / 100);
        palette[83 + i] = 'rgb(255,' + cc + ',0)';
      }
      // inner corona
      for (let i = 0; i < 201; i++) {
        let cc1 = 255 - Math.floor(i * 85 / 200);
        let cc2 = 200 - Math.floor(i * 30 / 200);
        let cc3 = Math.floor(i * 170 / 200);
        palette[184 + i] = 'rgb(' + cc1 + ',' + cc2 + ',' + cc3 + ')';
      }
      for (let i = 0; i < 301; i++) {
        let cc1 = 170 - Math.floor(i * 43 / 300);
        let cc2 = 170 + Math.floor(i * 85 / 300);
        palette[385 + i] = 'rgb(' + cc1 + ',' + cc1 + ',' + cc2 + ')';
      }
      for (let i = 0; i < 338; i++) {
        let cc = 127 + Math.floor(i * 128 / 337);
        palette[686 + i] = 'rgb(' + cc + ',' + cc + ',255)';
      }
      palette[1024] = 'rgb(0,0,0)';
      break;
    case 'rainbow-linear1':
      for (let i = 0; i < 256; i++) {
        palette[i] = 'rgb(' + i + ',0,0)';
        palette[256 + i] = 'rgb(255,' + i + ',0)';
        palette[512 + i] = 'rgb(' + (255 - i) + ',255,' + i + ')';
        palette[768 + i] = 'rgb(' + i + ',' + (255 - i) + ',255)';
      }
      palette[1024] = 'rgb(0,0,0)';
      break;
  }
/*
Select Case palnr
Case 1  'Standard-Palette (QB-Colors)
     For i = 0 To 1024
         xx = CInt(i * 500 / 1024 + 2)
         If xx <= 15 Then clr = xx
         If xx > 15 Then clr = CInt(Sqr((xx - 15 + 1) * 15 ^ 2 / 485))
         If xx >= 500 Then clr = 0
         palette(i) = QBColor(clr)
     Next
Case 3  'Regenbogen-Palette 1 (qu.)
     For i = 0 To 33
         clr = CInt(i * 255 / 33)
         palette(i) = RGB(clr, 0, 0)
     Next
     For i = 0 To 136
         clr = CInt(i * 255 / 136)
         palette(34 + i) = RGB(255, clr, 0)
     Next
     For i = 0 To 306
         clr = CInt(i * 255 / 306)
         palette(171 + i) = RGB(255 - clr, 255, clr)
     Next
     For i = 0 To 545
         clr = CInt(i * 255 / 545)
         palette(478 + i) = RGB(clr, 255 - clr, 255)
     Next
Case 4  'Regenbogen-Palette 2 (linear)
     For i = 0 To 204
         clr = CInt(i * 255 / 204)
         palette(i) = RGB(255, clr, 0)
         palette(204 + i) = RGB(255 - clr, 255, 0)
         palette(409 + i) = RGB(0, 255, clr)
         palette(614 + i) = RGB(0, 255 - clr, 255)
         palette(819 + i) = RGB(clr, 0, 255)
     Next
Case 5  'Regenbogen-Palette 2 (qu.)
     For i = 0 To 18
         clr = CInt(i * 255 / 18)
         palette(i) = RGB(255, clr, 0)
     Next
     For i = 0 To 73
         clr = CInt(i * 255 / 73)
         palette(20 + i) = RGB(255 - clr, 255, 0)
     Next
     For i = 0 To 167
         clr = CInt(i * 255 / 167)
         palette(93 + i) = RGB(0, 255, clr)
     Next
     For i = 0 To 297
         clr = CInt(i * 255 / 297)
         palette(261 + i) = RGB(0, 255 - clr, 255)
     Next
     For i = 0 To 464
         clr = CInt(i * 255 / 464)
         palette(559 + i) = RGB(clr, 0, 255)
     Next
*/
  return palette;
}

function drawPoint(context, img_x, img_y, C, iterMax, algorithm) {
  var itVal;
  switch (algorithm) {
    case 'oo':
      itVal = mandelbrotValueOO(C, iterMax);
      break;
    case 'numeric':
    default:
      itVal = mandelbrotValueNumeric(C, iterMax);
      break;
  }
  context.fillStyle = getColor(itVal, iterMax);
  context.fillRect (img_x, img_y, 1, 1); // x, y, width, height
}

/***** pure UI functions *****/

function saveImage() {
  // XXX: should call filepicker!
  saveCanvas(document.getElementById("mbrotImage"), "/home/robert/temp/canvas-save.png")
}

function updateIterMenu() {
  try {
    var currentIter = gPref.getIntPref("mandelbrot.iteration_max");
  }
  catch(e) {
    var currentIter = 0;
  }
  if (currentIter < 10) {
    currentIter = 500;
    setIter(currentIter);
  }

  var popup = document.getElementById("menu_iterPopup");
  var item = popup.firstChild;
  while (item) {
    if (item.getAttribute("name") == "iter") {
      if (item.getAttribute("value") == currentIter)
        item.setAttribute("checked","true");
      else
        item.removeAttribute("checked");
    }
    item = item.nextSibling;
  }
}

function setIter(aIter) {
  gPref.setIntPref("mandelbrot.iteration_max", aIter);
}

function updatePaletteMenu() {
  try {
    var currentPalette = gPref.getCharPref("mandelbrot.color_palette");
  }
  catch(e) {
    var currentPalette = '';
  }
  if (!currentPalette.length) {
    currentPalette = 'kairo';
    setPalette(currentPalette);
  }
  if (!gColorPalette || !gColorPalette.length)
    gColorPalette = getColorPalette(currentPalette);

  var popup = document.getElementById("menu_palettePopup");
  var item = popup.firstChild;
  while (item) {
    if (item.getAttribute("name") == "palette") {
      if (item.getAttribute("value") == currentPalette)
        item.setAttribute("checked", "true");
      else
        item.removeAttribute("checked");
    }
    item = item.nextSibling;
  }
}

function setPalette(aPaletteID) {
  gPref.setCharPref("mandelbrot.color_palette", aPaletteID);
  gColorPalette = getColorPalette(aPaletteID);
}

function updateDebugMenu() {
  var jitMenuItem = document.getElementById("jitEnabled");
  jitMenuItem.setAttribute("checked", gPref.getBoolPref("javascript.options.jit.chrome"));
}

function toggleJITState(jitMenuItem) {
  var jitEnabled = !gPref.getBoolPref("javascript.options.jit.chrome");
  gPref.setBoolPref("javascript.options.jit.chrome", jitEnabled)
  jitMenuItem.setAttribute("checked", jitEnabled? "true" : "false");
}

function updateAlgoMenu() {
  try {
    var currentAlgo = gPref.getCharPref("mandelbrot.use_algorithm");
  }
  catch(e) {
    var currentAlgo = '';
  }
  if (!currentAlgo.length) {
    currentAlgo = 'numeric';
    setAlgorithm(currentAlgo);
  }

  var popup = document.getElementById("menu_algoPopup");
  var item = popup.firstChild;
  while (item) {
    if (item.getAttribute("name") == "algorithm") {
      if (item.getAttribute("value") == currentAlgo)
        item.setAttribute("checked", "true");
      else
        item.removeAttribute("checked");
    }
    item = item.nextSibling;
  }
}

function setAlgorithm(algoID) {
  gPref.setCharPref("mandelbrot.use_algorithm", algoID);
}


/***** helper functions from external sources *****/

// function below is from from http://developer.mozilla.org/en/docs/Code_snippets:Canvas
function saveCanvas(canvas, destFile) {
  // convert string filepath to an nsIFile
  var file = Components.classes["@mozilla.org/file/local;1"]
                       .createInstance(Components.interfaces.nsILocalFile);
  file.initWithPath(destFile);

  // create a data url from the canvas and then create URIs of the source and targets  
  var io = Components.classes["@mozilla.org/network/io-service;1"]
                     .getService(Components.interfaces.nsIIOService);
  var source = io.newURI(canvas.toDataURL("image/png", ""), "UTF8", null);
  var target = io.newFileURI(file);

  // prepare to save the canvas data
  var persist = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"]
                          .createInstance(Components.interfaces.nsIWebBrowserPersist);

  persist.persistFlags = Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES;
  persist.persistFlags |= Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_AUTODETECT_APPLY_CONVERSION;

  // displays a download dialog (remove these 3 lines for silent download)
  var xfer = Components.classes["@mozilla.org/transfer;1"]
                       .createInstance(Components.interfaces.nsITransfer);
  xfer.init(source, target, "", null, null, null, persist);
  persist.progressListener = xfer;

  // save the canvas data to the file
  persist.saveURI(source, null, null, null, null, file);
}

// function below is from http://developer.mozilla.org/en/docs/How_to_Quit_a_XUL_Application
function quitApp(aForceQuit) {
  var appStartup = Components.classes['@mozilla.org/toolkit/app-startup;1']
                             .getService(Components.interfaces.nsIAppStartup);

  // eAttemptQuit will try to close each XUL window, but the XUL window can cancel the quit
  // process if there is unsaved data. eForceQuit will quit no matter what.
  var quitSeverity = aForceQuit ? Components.interfaces.nsIAppStartup.eForceQuit :
                                  Components.interfaces.nsIAppStartup.eAttemptQuit;
  appStartup.quit(quitSeverity);
}
